/* ----------------------------------------------------------------- */
/*  Copyright 1999 IBM Corporation.  All rights reserved.            */
/* ----------------------------------------------------------------- */

/**
 * File:             ServletBean.java
 * Date of Creation: 03/01/1999
 *   
 * Description: Servlet Admin Repository Object Bean implementation.
 *       
 * @author Michael J. Morton
 * @Modification for defect 88536. Delta from ASV302 1.19 (cmvc)
 */ 
package com.ibm.ejs.sm.beans;

import java.util.*;
import java.sql.*;
import java.rmi.RemoteException;
import javax.ejb.*;
import java.io.*;
import com.ibm.ejs.ras.*;
import com.ibm.ejs.sm.exception.*;
import com.ibm.ejs.sm.active.ActiveObjectConfig;
import com.ibm.ejs.sm.active.ActiveObjectStatus;
import com.ibm.ejs.sm.active.ActiveServletConfig;
import com.ibm.ejs.sm.agent.AdminAgent;
import com.ibm.ejs.sm.agent.ParamList;
import com.ibm.ejs.sm.agent.ContainmentPath;
import com.ibm.ejs.sm.agent.ContainmentPathElem;
import com.ibm.ejs.sm.client.ui.NLS;
import com.ibm.ejs.sm.util.act.Act;
import com.ibm.ejs.sm.util.act.ActServerImpl;
import com.ibm.ejs.sm.util.db.DBMgr;
import com.ibm.ejs.sm.util.db.DBQueryResult;
import com.ibm.ejs.sm.util.Utils;

public class ServletBean extends LiveRepositoryObjectImpl 
implements EntityBean 
{
	// Lock for synchronizing static variables.
	private static Integer classLock = new Integer(0);

	// Constant for Trace Component for this class.
	private static final TraceComponent tc = Tr.register(ServletBean.class);

	// Class/Constants for initialization of this class.
	private static NLS nls = null;
	protected static final String myInterfaceName = RepositoryObjectImpl.qualifyClassName("Servlet");
	private static final String myClassName = RepositoryObjectImpl.qualifyClassName("ServletBean");
	private static boolean tableCreated = false;
	private static Type myTypeObj = null;
	private static Long myTypeId = null;

	// Instance variables.
	private boolean dirty;
	private boolean hasContext = false;
	private ServletGroup initialServletGroup = null;

	// Instance variables for Servlet attributes.
	private String name;
	private String description;
	private String code;
	private Properties initParams;
	private boolean loadAtStartup;
	private boolean userServlet;
	private int debugMode;
	private boolean enabled;
	private Vector servletURIs; 

	/*
	 * Binary attribute names.
	 */
	private static final String initParamsBinAttrName = myInterfaceName +
																											".initParams";
	private static final String servletURIsBinAttrName = myInterfaceName +
																											 ".servletURIs";

	// Constants for default Servlet attribute values.
	private static final String defaultDescription = new String();
	private static final String defaultCode = new String();
	private static final Properties defaultInitParams = new Properties();
	private static final boolean defaultLoadAtStartup = false;
	private static final boolean defaultUserServlet = true;
	private static final int defaultDebugMode = 0;
	private static final boolean defaultEnabled = true;
	private static final Vector defaultServletURIs = new Vector();

	// Constants for location of attributes in Servlet Admin Table.
	private static final String tableName = DBMgr.qualifiedTableName("SRVLT_TBL"); 
	private static final String descriptionColumnName = "Description";
	private static final int descriptionColumnIndex = columnIndexBase + 1;
	private static final String codeColumnName = "Code";
	private static final int codeColumnIndex = columnIndexBase + 2;
	private static final String loadAtStartupColumnName = "LoadAtStartup";
	private static final int loadAtStartupColumnIndex = columnIndexBase + 3;
	private static final String userServletColumnName = "UserServlet";
	private static final int userServletColumnIndex = columnIndexBase + 4;
	private static final String debugModeColumnName = "DebugMode";
	private static final int debugModeColumnIndex = columnIndexBase + 5;
	private static final String enabledColumnName = "Enabled";
	private static final int enabledColumnIndex = columnIndexBase + 6;
	private static final int numColumns = columnIndexBase + 6;

	// Constants for index keys into prepared statement cache.
	private static final int servletBeanKeyBase = DBMgr.getPreparedStmtCacheKeyBase();
	private static final int loadStmtKey = servletBeanKeyBase + 1;   
	private static final int deleteStmtKey = servletBeanKeyBase + 2;
	private static final int updateStmtKey = servletBeanKeyBase + 3;
	private static final int findAllStmtKey = servletBeanKeyBase + 4;
	private static final int restrictedFindAllStmtKey = servletBeanKeyBase + 5;
	private static final int insertStmtKey = servletBeanKeyBase + 6;

	// Constants for Sql strings for manipulating the Servlet Admin Table.
	private static final String insertStmtSql = "insert into " + 
																							tableName   + 
																							" values(?,?,?,?,?,?,?,?)";

	private static final String updateStmtSql = "update " + 
																							tableName + 
																							" set " +
																							descriptionColumnName +
																							" = ? ," +
																							codeColumnName +
																							" = ? ," +
																							loadAtStartupColumnName  +
																							" =  ? ," +  
																							userServletColumnName  +
																							" =  ? ," +  
																							debugModeColumnName  +
																							" =  ? ," +  
																							enabledColumnName +
																							" =  ? " +
																							" where " + 
																							instanceIdColumnName +
																							" = ?"; 

	private static final String deleteStmtSql = "delete from " +
																							tableName + 
																							" where " + 
																							instanceIdColumnName +
																							" =  ?";

	// Static variables for information about relationships in which this
	// bean participates.
	private static Relation servletURIRel;
	protected static final String servletURIRelName = "servletURIRel";

	// Constant for active object class name corresponding to this bean.
	private static final String activeObjectClassName = "ActiveServlet";

	// Instance variables for related objects.
	private EJBObject containingObject;
	private ContainmentPathElem activeObjectName;
	private ServletGroup servletGroup;


	public ServletBean() throws RemoteException 
	{ 
		super();

		// Initialize for NLS.
		initializeNLS("AppServerROStrings");

		initializePersistentStore();
		initializeTypeId();
	}

	/*
	 * getAttributes -- Return the Attributes for this instance.
	 */
	public Attributes getAttributes(Attributes attributeNames) 
	throws RemoteException, OpException 
	{
		if (tc.isEntryEnabled()) Tr.entry(tc, "getAttributes");

		// Check to see if this instance is a sub-type of the Servlet bean.
		if (narrowRef != null)
		{
			return((Servlet)narrowRef).getAttributes(attributeNames);
		}

		if (tc.isEntryEnabled()) Tr.exit(tc, "getAttributes");
		return readInstanceVariables(attributeNames);
	}

	/*
	 * setAttributes - Set the Attributes for this instance.
	 */
	public void setAttributes(Attributes genericAttrs) 
	throws RemoteException, OpException  
	{
		if (tc.isEntryEnabled()) Tr.entry(tc, "setAttributes");

		if (narrowRef != null)
		{
			((Servlet)narrowRef).setAttributes(genericAttrs);
			if (tc.isEntryEnabled()) Tr.exit(tc, "setAttributes");
			return;
		}

		//mjm-55032
		renameIfNec(genericAttrs, ServletGroupBean.lookupGroupServletRel());

		try
		{
			Model model = isAClone();
			boolean isCloneInstance = false;

			if (model != null)
			{
				isCloneInstance = true;
			}
			updateInstanceVariables((ServletAttributes)genericAttrs, false, isCloneInstance);
		}
		catch (InvalidURINameException e)
		{
			RemoteException r = new RemoteException("", e);
			if (tc.isEntryEnabled()) Tr.exit(tc, "setAttributes");
			throw r;
		}

		if (tc.isEntryEnabled()) Tr.exit(tc, "setAttributes");
	}

	/*
	 * getActiveAttributes -- Get Servlet Active Attributes.
	 */
	private ServletAttributes getActiveAttributes(Attributes attributeNames) 
	throws RemoteException, OpException 
	{
		if (tc.isEntryEnabled()) Tr.entry(tc, "getActiveAttributes");

		ServletAttributes servletAttributes = (ServletAttributes) attributeNames;
		ActiveServletConfig config = null;

		if (getDesiredState() == ActiveObjectStatus.RUNNING)
		{
			try
			{
				config = (ActiveServletConfig) getRuntimeAttr("Config");
			}
			catch (Exception e)
			{
				if (tc.isEntryEnabled()) Tr.event(tc, "Exception occurred fetching config: ", e);
				config = null;
			}
		}

		if (config == null)
		{
			config = (ActiveServletConfig) getConfig(); 
		}

		servletAttributes.setDescriptionActive(config.getDescription());
		servletAttributes.setCodeActive(config.getCode());
		servletAttributes.setInitParamsActive(config.getInitParams());
		servletAttributes.setLoadAtStartupActive(config.getLoadAtStartup());
		servletAttributes.setUserServletActive(config.getUserServlet());
		servletAttributes.setEnabledActive(config.getEnabled());
		servletAttributes.setURIPathsActive(config.getURIPaths());

		if (tc.isEntryEnabled()) Tr.exit(tc, "getActiveAttributes");
		return servletAttributes;
	}

	/*
	 * initializePersistentStore -- Create Servlet table in DB if necessary.
	 */
	private static void initializePersistentStore() 
	throws RemoteException 
	{
		if (tc.isEntryEnabled()) Tr.entry(tc, "initializePersistentStore");

		synchronized (classLock)
		{
			if (tableCreated)
			{
				if (tc.isEntryEnabled()) Tr.exit(tc, "initializePersistentStore");
				return;
			}
		}

		try
		{
			String createTableSql =  "create table " + tableName + " ( " + 
															 instanceIdColumnName + DBMgr.idColumnSpec + " , " +
															 typeIdColumnName + DBMgr.idColumnSpec + " , " +  
															 descriptionColumnName + DBMgr.nullableMediumStringColumnSpec +  " , " + 
															 codeColumnName + DBMgr.mediumStringColumnSpec +  " , " + 
															 loadAtStartupColumnName + " INTEGER NOT NULL , " +
															 userServletColumnName + " INTEGER NOT NULL , " +
															 debugModeColumnName + " INTEGER NOT NULL , " +
															 enabledColumnName + " INTEGER NOT NULL , " +
															 " primary key(" +  instanceIdColumnName + "))";

			DBMgr.createTableIfNonExistent(tableName, createTableSql);

			synchronized (classLock)
			{
				tableCreated = true;
			}
		}
		catch (SQLException s)
		{
			RemoteException r = new RemoteException((nls.getString("sb.db.create.exception",
																														 "Database create table failed.")), 
																							s); 
			if (tc.isEntryEnabled()) Tr.exit(tc, "initializePersistentStore", r);
			throw r;
		}

		if (tc.isEntryEnabled()) Tr.exit(tc, "initializePersistentStore");
	}

	/*
	 * initializeTypeID -- Registration with the type bean to obtain a type identifier.
	 */
	private static void initializeTypeId() 
	throws RemoteException 
	{
		if (tc.isEntryEnabled()) Tr.entry(tc, "initializeTypeId");

		synchronized (classLock) 
		{
			if (myTypeId != null)
			{
				if (tc.isEntryEnabled()) Tr.exit(tc, "initializeTypeId");
				return;
			}
		}

		ServletAttributes defaultAttrs = new ServletAttributes();

		defaultAttrs.setDescription(defaultDescription);
		defaultAttrs.setCode(defaultCode);
		defaultAttrs.setInitParams(defaultInitParams);
		defaultAttrs.setLoadAtStartup(defaultLoadAtStartup);
		defaultAttrs.setUserServlet(defaultUserServlet);
		defaultAttrs.setDebugMode(defaultDebugMode);
		defaultAttrs.setEnabled(defaultEnabled);
		defaultAttrs.setURIPaths(defaultServletURIs);

		Type typeObj = RepositoryObjectImpl.getTypeObj(myInterfaceName, myClassName, defaultAttrs, false);
		Long typeId = (Long)typeObj.getPrimaryKey();

		synchronized (classLock) 
		{
			myTypeObj = typeObj;
			myTypeId = typeId;
		}

		if (tc.isEntryEnabled()) Tr.exit(tc, "initializeTypeId");
	}

	/*
	 * Methods mandated by the entity bean interface.
	 */

	/*
	 * ejbLoad -- Read from DB using primary key.
	 */
	public void ejbLoad()
	throws RemoteException 
	{
		if (tc.isEntryEnabled()) Tr.entry(tc, "ejbLoad");

		Long primaryKey = (Long) ec.getPrimaryKey();
		DBQueryResult qs = null;

		try
		{
			qs = DBMgr.executePreparedReadByPrimaryKey(loadStmtKey,
																								 tableName,
																								 instanceIdColumnName, 
																								 Utils.getIdString(primaryKey));
			ResultSet rs = qs.getResultSet();
			if (!rs.next())
			{
				RemoteException r = new RemoteException((nls.getFormattedMessage("sb.db.select1.exception",
																																				 new Object[] { primaryKey},
																																				 "Database select failed to find Servlet instance (id={0}).")),
																								new ObjectNotFoundException());
				if (tc.isEntryEnabled()) Tr.exit(tc, "ejbLoad", r);
				throw r;
			}

			if (setId(rs, myTypeId))
			{
				super.ejbLoad();

				description = rs.getString(descriptionColumnIndex);
				if (description == null)
				{
					description = new String();
				}
				code = rs.getString(codeColumnIndex);
				initParams = (Properties)getBinaryAttr(initParamsBinAttrName);
				loadAtStartup = Utils.intToBoolean(rs.getInt(loadAtStartupColumnIndex));
				userServlet = Utils.intToBoolean(rs.getInt(userServletColumnIndex));
				debugMode = rs.getInt(debugModeColumnIndex);
				enabled = Utils.intToBoolean(rs.getInt(enabledColumnIndex));
				servletURIs = (Vector)getBinaryAttr(servletURIsBinAttrName);
			}

			// Reset local variables for related objects.
			containingObject = null;
			activeObjectName = null;

			// Attribute values are current.
			dirty = false;
			hasContext = true;
		}
		catch (SQLException s)
		{
			RemoteException r = new RemoteException((nls.getString("sb.db.select2.exception",
																														 "Database select failed.")), 
																							s);
			if (tc.isEntryEnabled()) Tr.exit(tc, "ejbLoad", r);
			throw r;
		}
		finally
		{
			DBMgr.doneWithQuery(qs);
		}

		if (tc.isEntryEnabled()) Tr.exit(tc, "ejbLoad");
	}

	/*
	 * ejbStore -- Invoke the super class and then update instance
	 *             record in DB if dirty.
	 */
	public void ejbStore()
	throws RemoteException 
	{
		if (tc.isEntryEnabled()) Tr.entry(tc, "ejbStore");

		if (narrowRef != null)
		{
			if (tc.isEntryEnabled()) Tr.exit(tc, "ejbStore");
			return;
		}

		super.ejbStore();

		if (dirty)
		{
			try
			{
				Vector params = new Vector(numColumns);

				params.addElement(description);
				params.addElement(code);
				params.addElement(new Integer(Utils.booleanToInt(loadAtStartup)));
				params.addElement(new Integer(Utils.booleanToInt(userServlet)));
				params.addElement(new Integer(debugMode));
				params.addElement(new Integer(Utils.booleanToInt(enabled)));
				params.addElement(Utils.getIdString(id));

				DBMgr.executePreparedUpdate(updateStmtKey, updateStmtSql, params);
				dirty = false;
			}
			catch (SQLException e)
			{
				RemoteException r = new RemoteException((nls.getFormattedMessage("sb.db.update.exception",
																																				 new Object[] { id},
																																				 "Database update failed to store Servlet attributes (id={0}).")), 
																								e); 
				if (tc.isEntryEnabled()) Tr.exit(tc, "ejbStore", r);
				throw r;
			}
		}

		if (tc.isEntryEnabled()) Tr.exit(tc, "ejbStore");
	}

	/*
	 * ejbRemove -- Delete the instance entry from table after calling
	 *              superclass.
	 */
	public void ejbRemove() 
	throws RemoteException, RemoveException 
	{
		if (tc.isEntryEnabled()) Tr.entry(tc, "ejbRemove");

		Long primaryKey = (Long) ec.getPrimaryKey();

		if (id == null)
		{
			ejbLoad();
		}

		if (narrowRef != null)
		{
			narrowRef.remove(); 
			if (tc.isEntryEnabled()) Tr.exit(tc, "ejbRemove");
			return;
		}

		// Need to determine if this instance is a Clone.
		boolean isClone = true;
		try
		{
			Model model = isAClone();
			if (model == null)
			{
				isClone = false;
			}
			else
			{
				isClone = true;
            ServletBeanHelper.reSet();//88536
			}
		}
		catch (OpException e)
		{
			// When in doubt, don't delete the URIs!
			isClone = true;
		}

		super.ejbRemove();

		try
		{
			Vector params = new Vector(1);
			params.addElement(Utils.getIdString(primaryKey));

			DBMgr.executePreparedUpdate(deleteStmtKey, deleteStmtSql, params); 
		}
		catch (SQLException s)
		{
			ec.setRollbackOnly();
			RemoteException r = new RemoteException((nls.getFormattedMessage("sb.db.remove.exception",
																																			 new Object[] { primaryKey},
																																			 "Database delete failed to remove Servlet instance (id={0}).")), 
																							s);
			if (tc.isEntryEnabled()) Tr.exit(tc, "ejbRemove", r);
			throw r;
		}

		// If this Servlet is not a Clone, then remove the URI instances associated with it.
		if (!isClone)
		{
			URI uri = null;
			URIHome uriHome = (URIHome)javax.rmi.PortableRemoteObject.narrow(getHome("URIHome"), URIHome.class);
			int numURIs = servletURIs.size();

			for (int i = 0; i < numURIs; i++)
			{
				try
				{
					uri = uriHome.findByName((String)(servletURIs.elementAt(i)), false);
					uri.setPrivateRemove(new Boolean(true));
					uri.remove();
				}
				catch (Exception e)
				{
				}
			}
		}

		if (tc.isEntryEnabled()) Tr.exit(tc, "ejbRemove");
	}

	/*
	 * ejbCreate -- Create an instance entry in the DB table.
	 */
	public Long ejbCreate(ServletAttributes servletAttributes, ServletGroup group)
	throws RemoteException, CreateException, AttributeNotSetException 
	{
		if (tc.isEntryEnabled()) Tr.entry(tc, "ejbCreate");

		// Create of new Servlet has no Context yet.
		hasContext = false;

		try
		{
			name = servletAttributes.getName();
			servletAttributes.getCode();
			servletAttributes.getURIPaths();
		}
		catch (AttributeNotSetException e)
		{
			RequiredAttributeMissingException r = new RequiredAttributeMissingException(nls.getString("sb.attribute.missing1.exception",
																																																"Required Servlet name or code attribute was not specified."));
			if (tc.isEntryEnabled()) Tr.exit(tc,"ejbCreate -- required attribute missing", r);
			throw r;
		}

		try
		{
			// Check whether we are being invoked from a sub-class. 
			// If not, we are actually creating an instance of "this type".
			if (typeId == null)
			{
				typeId = myTypeId;
			}

			id = create(typeId); 

			super.ejbCreate(); 

			// Need access to input ServletGroup. 
			initialServletGroup = group;

			// Allocate id and initialize instance variables. 
			initializeInstanceVariables();
			updateInstanceVariables(servletAttributes, false, false);

			// Add ServletGroup-Servlet relation. 
			Relation rel = ServletGroupBean.lookupGroupServletRel();

			try
			{
				rel.addUsingKeys((Long)group.getPrimaryKey(), id, name);
			}
			catch (RelationOpException e)
			{
				InvalidServletNameException i = new InvalidServletNameException(nls.getFormattedMessage("sb.already.exists.exception",
																																																new Object[] { name},
																																																"Servlet already exists (name={0})."));
				if (tc.isEntryEnabled()) Tr.exit(tc, "ejbCreate", i);
				throw i;
			}

			Vector params = new Vector(numColumns);    
			params.addElement(Utils.getIdString(id));
			params.addElement(Utils.getIdString(typeId));
			params.addElement(description);
			params.addElement(code);
			params.addElement(new Integer(Utils.booleanToInt(loadAtStartup)));
			params.addElement(new Integer(Utils.booleanToInt(userServlet)));
			params.addElement(new Integer(debugMode));
			params.addElement(new Integer(Utils.booleanToInt(enabled)));

			DBMgr.executePreparedUpdate(insertStmtKey, insertStmtSql, params);
		}
		catch (SQLException s)
		{
			RemoteException r = new RemoteException((nls.getFormattedMessage("sb.db.insert.exception",
																																			 new Object[] { id},
																																			 "Database insert failed to store new Servlet instance (id={0}).")), 
																							s);
			if (tc.isEntryEnabled()) Tr.exit(tc, "ejbCreate", r);
			throw r;
		}

		if (tc.isEntryEnabled()) Tr.exit(tc, "ejbCreate", id);
		return id;
	}

	/*
	 * ejbCreate -- Create an instance entry in the DB table.
	 */
	public Long ejbCreate(Model model, ServletAttributes servletAttributes, ServletGroup group)
	throws RemoteException, CreateException, AttributeNotSetException 
	{
		if (tc.isEntryEnabled()) Tr.entry(tc, "ejbCreate");

		// Create of new Servlet has no Context yet.
		hasContext = false;

		try
		{
			name = servletAttributes.getName();
			servletAttributes.getCode();
			servletAttributes.getURIPaths();
		}
		catch (AttributeNotSetException e)
		{
			RequiredAttributeMissingException r = new RequiredAttributeMissingException(nls.getString("sb.attribute.missing1.exception",
																																																"Required Servlet name or code attribute was not specified."));
			if (tc.isEntryEnabled()) Tr.exit(tc,"ejbCreate -- required attribute missing", r);
			throw r;
		}

		try
		{
			// Check whether we are being invoked from a sub-class. 
			// If not, we are actually creating an instance of "this type".
			if (typeId == null)
			{
				typeId = myTypeId;
			}

			id = create(typeId); 

			super.ejbCreate(); 

			// Need access to input ServletGroup. 
			initialServletGroup = group;

			// Allocate< id and initialize instance variables. 
			initializeInstanceVariables();
			updateInstanceVariables(servletAttributes, true, false);

			// Add ServletGroup-Servlet relation. 
			Relation rel = ServletGroupBean.lookupGroupServletRel();

			try
			{
				rel.addUsingKeys((Long)group.getPrimaryKey(), id, name);
			}
			catch (RelationOpException e)
			{
				InvalidServletNameException i = new InvalidServletNameException(nls.getFormattedMessage("sb.already.exists.exception",
																																																new Object[] { name},
																																																"Servlet already exists (name={0})."));
				if (tc.isEntryEnabled()) Tr.exit(tc, "ejbCreate", i);
				throw i;
			}

			Vector params = new Vector(numColumns);    
			params.addElement(Utils.getIdString(id));
			params.addElement(Utils.getIdString(typeId));
			params.addElement(description);
			params.addElement(code);
			params.addElement(new Integer(Utils.booleanToInt(loadAtStartup)));
			params.addElement(new Integer(Utils.booleanToInt(userServlet)));
			params.addElement(new Integer(debugMode));
			params.addElement(new Integer(Utils.booleanToInt(enabled)));

			DBMgr.executePreparedUpdate(insertStmtKey, insertStmtSql, params);
         ServletBeanHelper.reSet();//88536
		}
		catch (SQLException s)
		{
			RemoteException r = new RemoteException((nls.getFormattedMessage("sb.db.insert.exception",
																																			 new Object[] { id},
																																			 "Database insert failed to store new Servlet instance (id={0}).")), 
																							s);
			if (tc.isEntryEnabled()) Tr.exit(tc, "ejbCreate", r);
			throw r;
		}

		if (tc.isEntryEnabled()) Tr.exit(tc, "ejbCreate", id);
		return id;
	}

	/*
	 * ejbPostCreate -- Post ejbCreate() opportunity.
	 */
	public void ejbPostCreate(ServletAttributes servletAttributes, ServletGroup group)
	{
	} 

	public void ejbPostCreate(Model model, ServletAttributes servletAttributes, ServletGroup group)
	{
	} 

	/*
	 * Finder methods
	 */

	/*
	 * ejbFindAll -- Return enumeration of all DB table entries.
	 */
	public Enumeration ejbFindAll(boolean useSubtypes) 
	throws RemoteException, FinderException 
	{
		if (tc.isEntryEnabled()) Tr.entry(tc, "ejbFindAll");

		int stmtKey;

		if (useSubtypes)
		{
			stmtKey = findAllStmtKey;
		}
		else
		{
			stmtKey = restrictedFindAllStmtKey;
		}

		Enumeration keyList;
		try
		{
			keyList = DBMgr.executeFindAll(stmtKey, 
																		 tableName,
																		 useSubtypes, 
																		 typeIdColumnName,
																		 myTypeId, 
																		 instanceIdColumnName);
		}
		catch (SQLException s)
		{
			RemoteException r = new RemoteException((nls.getString("sb.db.select3.exception",
																														 "Database select failed to find all instances of Servlet.")), 
																							s);
			if (tc.isEntryEnabled()) Tr.exit(tc, "ejbFindAll", r);
			throw r;
		}

		if (tc.isEntryEnabled()) Tr.exit(tc, "ejbFindAll", keyList);
		return keyList;
	}

	/*
	 * initializeInstanceVariables -- Initialize instance variables.
	 */
	private void initializeInstanceVariables() 
	throws RemoteException 
	{
		if (tc.isEntryEnabled()) Tr.entry(tc, "initializeInstanceVariables");

		try
		{
			ServletAttributes defaultAttrs = (ServletAttributes) RepositoryObjectImpl.getDefaults(myTypeObj);

			description = defaultAttrs.getDescription();
			code = defaultAttrs.getCode();
			initParams = defaultAttrs.getInitParams();
			setBinaryAttr(initParamsBinAttrName, initParams);
			loadAtStartup = defaultAttrs.getLoadAtStartup();
			userServlet = defaultAttrs.getUserServlet();
			debugMode = defaultAttrs.getDebugMode();
			enabled = defaultAttrs.getEnabled();
			servletURIs = defaultAttrs.getURIPaths();
			setBinaryAttr(servletURIsBinAttrName, servletURIs);
		}
		catch (Exception e)
		{
			RemoteException r = new RemoteException((nls.getString("sb.attribute.init.exception",
																														 "Exception occurred initializing default Servlet attributes.")), 
																							e);
			if (tc.isEntryEnabled()) Tr.exit(tc, "initializeInstanceVariables", r);
			throw r;
		}

		// Reset local variables for related objects.
		containingObject = null;
		activeObjectName = null;

		if (tc.isEntryEnabled()) Tr.exit(tc, "initializeInstanceVariables");
	}

	/*
	 * updateInstanceVariables -- Update instance variables.
	 */
	private void updateInstanceVariables(ServletAttributes servletAttributes, boolean isCloneCreate, boolean isCloneInstance)
	throws RemoteException, InvalidURINameException 
	{
		if (tc.isEntryEnabled()) Tr.entry(tc, "updateInstanceVariables");

		dirty = true;

		try
		{
			if (servletAttributes.isSet(ServletAttributes.description))
				description = servletAttributes.getDescription();
			if (servletAttributes.isSet(ServletAttributes.code))
				code = servletAttributes.getCode();
			if (servletAttributes.isSet(ServletAttributes.initParams))
			{
				initParams = servletAttributes.getInitParams();
				setBinaryAttr(initParamsBinAttrName, initParams);
			}
			if (servletAttributes.isSet(ServletAttributes.loadAtStartup))
				loadAtStartup = servletAttributes.getLoadAtStartup();
			if (servletAttributes.isSet(ServletAttributes.userServlet))
				userServlet = servletAttributes.getUserServlet();
			if (servletAttributes.isSet(ServletAttributes.debugMode))
				debugMode = servletAttributes.getDebugMode();
			if (servletAttributes.isSet(ServletAttributes.enabled))
				enabled = servletAttributes.getEnabled();
			if (servletAttributes.isSet(ServletAttributes.uriPaths))
			{
				if (isCloneCreate)
				{
					servletURIs = servletAttributes.getURIPaths();
               try{
               setURIs(servletAttributes.getURIPaths(), isCloneInstance);
               }catch(com.ibm.ejs.sm.exception.InvalidURINameException ie){
                  //ignore it
               }
					setBinaryAttr(servletURIsBinAttrName, servletURIs);
				}
				else
				{
					setURIs(servletAttributes.getURIPaths(), isCloneInstance);
					setBinaryAttr(servletURIsBinAttrName, servletURIs);
				}
			}
		}
		catch (AttributeNotSetException e)
		{
			RemoteException r = new RemoteException((nls.getString("sb.attribute.missing2.exception",
																														 "Required attribute for Servlet was not input.")), 
																							e);
			if (tc.isEntryEnabled()) Tr.exit(tc, "updateInstanceVariables", r);
			throw r;
		}
		catch (AttributeDoesNotExistException e)
		{
			RemoteException r = new RemoteException((nls.getString("sb.attribute.not.exist.exception",
																														 "Input attribute does not exist for a Servlet.")), 
																							e);
			if (tc.isEntryEnabled()) Tr.exit(tc, "updateInstanceVariables", r);
			throw r;
		}

		if (tc.isEntryEnabled()) Tr.exit(tc, "updateInstanceVariables");
	}

	/*
	 * readInstanceVariables -- Return current instance variables.
	 */
	private ServletAttributes readInstanceVariables(Attributes attributeNames) 
	throws RemoteException, OpException 
	{ 
		if (tc.isEntryEnabled()) Tr.entry(tc, "readInstanceVariables");

		ServletAttributes servletAttributes = (ServletAttributes) attributeNames;

		servletAttributes = (ServletAttributes) super.getAttributes(servletAttributes);
		servletAttributes.setName(getRelativeName());
		servletAttributes.setDescription(description);
		servletAttributes.setCode(code);
		servletAttributes.setInitParams(initParams);
		servletAttributes.setLoadAtStartup(loadAtStartup);
		servletAttributes.setUserServlet(userServlet);
		servletAttributes.setDebugMode(debugMode);
		servletAttributes.setEnabled(enabled);
		servletAttributes.setURIPaths(servletURIs);

		if (servletAttributes.activeAttributeRequested())
		{
			getActiveAttributes(attributeNames);
		}

		if (tc.isEntryEnabled()) Tr.exit(tc, "readInstanceVariables");
		return servletAttributes;
	}

	/*
	 * Relationship Methods
	 */

	/* 
	 * getRelativeName -- Get the name from the Relation.
	 */
	protected String getRelativeName() 
	throws RemoteException 
	{
		return super.getName(ServletGroupBean.lookupGroupServletRel());
	}

	/*
	 * getFullName -- Get the fully-qualified name.
	 */
	public RepositoryObjectName getFullName()
	throws RemoteException, OpException
	{
		return constructFullName("ServletHome", getRelativeName());
	}

	/*
	 * getContainingObject -- Get ServletGroup that contains this Servlet.
	 */ 
	public EJBObject getContainingObject() 
	throws RemoteException, ObjectNotFoundException 
	{
		if (tc.isEntryEnabled()) Tr.entry(tc, "getContainingObject");

		if (containingObject != null)
		{
			if (tc.isEntryEnabled()) Tr.exit(tc, "getContainingObject");
			return containingObject;
		}

		Long primaryKey = (Long) ec.getPrimaryKey();

		Relation rel = ServletGroupBean.lookupGroupServletRel();
		ServletGroup servletGroup;

		try
		{
			servletGroup = (ServletGroup)javax.rmi.PortableRemoteObject.narrow(rel.traverse(ec.getEJBObject(), true), ServletGroup.class);
		}
		catch (RelationOpException e)
		{
			RemoteException r = new RemoteException((nls.getFormattedMessage("sb.relation.traverse.exception",
																																			 new Object[] { primaryKey},
																																			 "Exception occurred while traversing Relation instances looking for ServletGroup for the Servlet (id={0}).")), 
																							e);
			if (tc.isEntryEnabled()) Tr.exit(tc, "getContainingObject", r);
			throw r;
		}

		containingObject = servletGroup;

		if (tc.isEntryEnabled()) Tr.exit(tc, "getContainingObject");
		return servletGroup;
	}

	/*
	 * listServletURIObjects -- Return vector of URIs contained by this
	 *                          Servlet.
	 */
	public Vector listServletURIObjects() 
	throws RemoteException 
	{
		if (tc.isEntryEnabled()) Tr.entry(tc, "listServletURIObjects");

		URIHome uriHome = (URIHome)javax.rmi.PortableRemoteObject.narrow(getHome("URIHome"), URIHome.class);
		URI uri = null;

		Vector uris = new Vector();
		int numURIs = servletURIs.size();

		for (int i = 0; i < numURIs; i++)
		{
			try
			{
				uri = uriHome.findByName((String)(servletURIs.elementAt(i)), false);
				uris.addElement(uri);
			}
			catch (FinderException e)
			{
				e.printStackTrace();
				Long primaryKey = (Long) ec.getPrimaryKey();
				RemoteException r = new RemoteException((nls.getFormattedMessage("sb.servlet.uri.find.exception",
																																				 new Object[] { primaryKey},
																																				 "Exception occurred attempting to get the URI for the Servlet (id={0}).")),
																								e);
				if (tc.isEntryEnabled()) Tr.exit(tc, "listServletURIObjects", r);
				throw r;
			}
		}

		if (tc.isEntryEnabled()) Tr.exit(tc, "listServletURIObjects");
		return uris;
	}

	/*
	 * ActiveObject Methods
	 */

	/*
	 * getActiveObjectName -- Return the Active Object for this Servlet instance.
	 */
	public ContainmentPathElem getActiveObjectName() 
	throws RemoteException, OpException 
	{
		if (tc.isEntryEnabled()) Tr.entry(tc, "getActiveObjectName");


		if (activeObjectName != null)
		{
			if (tc.isEntryEnabled()) Tr.exit(tc, "getActiveObjectName");
			return activeObjectName;
		}

		activeObjectName = new ContainmentPathElem(id,
																							 activeObjectClassName,
																							 getRelativeName(),
																							 getEpoch());

		if (tc.isEntryEnabled()) Tr.exit(tc, "getActiveObjectName");
		return activeObjectName;
	}

	/*
	 * getConfig -- Return the the ActiveServlet config. 
	 */
	public ActiveObjectConfig getConfig() 
	throws RemoteException, OpException 
	{
		if (tc.isEntryEnabled()) Tr.entry(tc, "getConfig");

		ActiveServletConfig config = new ActiveServletConfig();

		super.initializeConfig(config);

		config.setDescription(description);
		config.setCode(code);
		config.setInitParams(initParams);
		config.setLoadAtStartup(loadAtStartup);
		config.setUserServlet(userServlet);
		config.setDebugMode(debugMode);
		config.setEnabled(enabled);
		config.setURIPaths(servletURIs);

		if (tc.isEntryEnabled()) Tr.exit(tc, "getConfig");
		return config;
	}

	/*
	 * getActiveServletConfig() -- Returns a current ActiveObjectConfig
	 *                             object for this servlet.
	 */
	public ActiveObjectConfig getActiveServletConfig()
	throws RemoteException, OpException
	{
		if (tc.isEntryEnabled()) Tr.entry(tc, "getActiveObjectConfig");

		ActiveObjectConfig config = getConfig();

		if (tc.isEntryEnabled()) Tr.exit(tc, "getActiveObjectConfig");
		return config;
	}

	/*
	 * setEnabled -- Schedules an After-Commit-Task for invoking
	 *               the setEnabled() method of ActiveServlet.
	 */
	public Act setEnabled(Boolean flag)
	throws RemoteException, OpException
	{
		if (tc.isEntryEnabled()) Tr.entry(tc, "setEnabled");

		String activeMethod = "setEnabledAction";
		boolean enabledFlag = flag.booleanValue();

		try
		{
			if ((enabled != enabledFlag) && (getDesiredState() == ActiveObjectStatus.RUNNING))
			{
				AdminAgent nodeAgent = getNode().getAdminAgent();
				ContainmentPath target = getActiveObjectContainmentPath();
				ActServerImpl actServer = ActServerImpl.getInstance();   

				ServletGroup servletGroup = null;
				String groupName = null;

				try
				{
					servletGroup = (ServletGroup) getContainingObject();
					ServletGroupAttributes groupAttrs = new ServletGroupAttributes();
					groupAttrs = (ServletGroupAttributes) (servletGroup.getAttributes(groupAttrs));
					groupName = groupAttrs.getName();
				}
				catch (ObjectNotFoundException e)
				{
					ServletSetEnabledException s = new ServletSetEnabledException(nls.getString("sb.servlet.exception",
																																											"Servlet does not belong to a ServletGroup."));
					e.printStackTrace();
					if (tc.isEntryEnabled()) Tr.exit(tc, "setEnabled");
					throw s;
				}

				VirtualHost vhost = servletGroup.getVirtualHost();
				String vhostName = null;

				if (vhost != null)
				{
					VirtualHostAttributes vhostAttrs = new VirtualHostAttributes();
					vhostAttrs = (VirtualHostAttributes) (vhost.getAttributes(vhostAttrs));
					vhostName = vhostAttrs.getName();
				}

				int epoch = super.getEpoch();
				int version = super.getVersion();

				ParamList paramList = new ParamList(5);
				paramList.addElement(enabledFlag);
				paramList.addElement(vhostName);
				paramList.addElement(groupName);
				paramList.addElement(epoch);
				paramList.addElement(version);

				// Indicates to ejbStore() to persist the enabled state.
				enabled = enabledFlag;
				dirty = true;

				Act act = actServer.scheduleWork(new ActionTask(nodeAgent, 
																												target,
																												activeMethod,
																												paramList,
																												(EJBObject)ec.getEJBObject()));

				if (tc.isEntryEnabled()) Tr.exit(tc, "setEnabled");
				return act;
			}
		}
		catch (OpException e)
		{
			ServletSetEnabledException s = new ServletSetEnabledException(nls.getFormattedMessage("sb.act.schedule.exception",
																																														new Object[] { activeMethod},
																																														"Exception occurred scheduling After-Commit-Task (method={0})."));
			e.printStackTrace();
			if (tc.isEntryEnabled()) Tr.exit(tc, "setEnabled", s);
			throw s;                                                                  
		}
		catch (RemoteException e)
		{
			ServletSetEnabledException s = new ServletSetEnabledException(nls.getFormattedMessage("sb.act.schedule.exception",
																																														new Object[] { activeMethod},
																																														"Exception occurred scheduling After-Commit-Task (method={0})."));
			e.printStackTrace();
			if (tc.isEntryEnabled()) Tr.exit(tc, "setEnabled", s);
			throw s;
		}
		catch (AttributeNotSetException e)
		{
			ServletSetEnabledException s = new ServletSetEnabledException(nls.getString("sb.attribute.not.set1.exception",
																																									"VirtualHost name attribute was not set."));
			e.printStackTrace();
			if (tc.isEntryEnabled()) Tr.exit(tc, "setEnabled", s);
			throw s;
		}

		if (tc.isEntryEnabled()) Tr.exit(tc, "setEnabled - Servlet is not RUNNING, returning null.");
		return null;
	}

	/*
	 * setRefreshConfig - Invokes the ActiveServlet setRefreshConfig() method
	 *                    to initiate an action on this Servlet.
	 */
	public Act setRefreshConfig(Integer rAction)
	throws RemoteException, OpException
	{
		if (tc.isEntryEnabled()) Tr.entry(tc, "setRefreshConfig");

		String activeMethod = "setRefreshConfigAction";
		int refreshAction = rAction.intValue();

		try
		{
			if (getDesiredState() == ActiveObjectStatus.RUNNING)
			{
				AdminAgent nodeAgent = getNode().getAdminAgent();
				ContainmentPath target = getActiveObjectContainmentPath();
				ActServerImpl actServer = ActServerImpl.getInstance();   

				ActiveObjectConfig config = getConfig();

				int epoch = super.getEpoch();
				int version = super.getVersion();

				ParamList paramList = new ParamList(4);
				paramList.addElement(refreshAction);
				paramList.addElement(config);
				paramList.addElement(epoch);
				paramList.addElement(version);

				Act act = actServer.scheduleWork(new ActionTask(nodeAgent, 
																												target,
																												activeMethod,
																												paramList,
																												(EJBObject)ec.getEJBObject()));

				if (tc.isEntryEnabled()) Tr.exit(tc, "setRefreshConfig");
				return act;
			}
		}
		catch (OpException e)
		{
			ServletSetRefreshConfigException s = new ServletSetRefreshConfigException(nls.getFormattedMessage("sb.act.schedule.exception",
																																																				new Object[] { activeMethod},
																																																				"Exception occurred scheduling After-Commit-Task (method={0})."));
			e.printStackTrace();
			if (tc.isEntryEnabled()) Tr.exit(tc, "setRefreshConfig", s);
			throw s;
		}
		catch (RemoteException e)
		{
			ServletSetRefreshConfigException s = new ServletSetRefreshConfigException(nls.getFormattedMessage("sb.act.schedule.exception",
																																																				new Object[] { activeMethod},
																																																				"Exception occurred scheduling After-Commit-Task (method={0})."));
			e.printStackTrace();
			if (tc.isEntryEnabled()) Tr.exit(tc, "setRefreshConfig", s);
			throw s;
		}

		if (tc.isEntryEnabled()) Tr.exit(tc, "setRefreshConfig - Servlet is not RUNNING, returning null.");
		return null;
	}

	/*
	 * getURINames - Returns a Vector of the URI names for the URIs
	 *               that this Servlet is associated with. The format
	 *               of the URI name is:
	 *                   <virtualhost><root_uri><uri><uri>...
	 */
	public Vector getURINames() 
	throws RemoteException
	{
		if (tc.isEntryEnabled()) Tr.entry(tc, "getURINames");
		return servletURIs;
	}

	/*
	 * getURIs - Returns the URIs associated with this Servlet, if any exists.
	 */
	public Vector getURIs()
	throws RemoteException
	{
		return listServletURIObjects();
	}

	/*
	 * setURIs - Sets the Servlet URIs attribute and creates the corresponding URI 
	 *           instances for the input vector of URI paths. The input URI path format is:
	 *              <virtualhost><root_uri><uri>...
	 */
	public void setURIs(Vector newURIPaths, boolean isCloneInstance)
	throws RemoteException, InvalidURINameException
	{
		if (tc.isEntryEnabled()) Tr.entry(tc, "setURIs"); 
		Long uriKey = null;

		try
		{
			Long primaryKey = null;
         //Long pk_clone=new Long(0); //@88536A
         boolean deleteFlag=false;//@88563A
			Vector currentURIs = null;
			URI uri = null;
			URI currentURI = null;
			int currentLen = 0;
			int newLen = 0;
			int j = 0;
			boolean found = false;
			Vector uris = null;
			Vector uriPaths = null;
			Servlet servlet = null;
			Vector servletList = null;
			ServletGroup servletGroup = null;
			Vector groupList = null;
			int listSize = 0;

			// If no URI paths were input, then throw exception.
			if (newURIPaths == null)
			{
				RemoteException r = new RemoteException(nls.getString("sb.seturis.null.uri.exception",
																															"No URI path(s) were input for the Servlet."));
				if (tc.isEntryEnabled()) Tr.exit(tc, "setURIs", r);
				throw r;
			}

			URIHome uriHome = (URIHome)javax.rmi.PortableRemoteObject.narrow(getHome("URIHome"), URIHome.class);

			int size = newURIPaths.size();
			uris = new Vector(size);
			uriPaths = new Vector(size);
			String uriString = null;
			String groupPath = null;
			URI newURI = null;

			// Given the input Vector of URI paths, attempt to find the URI by name. 
			// If it does not exist, then create one with the given path.            
			for (int i = 0; i < size; i++)
			{
				uriString = newURIPaths.elementAt(i).toString();
				uriString = URIBean.removeDoubleSlash(uriString);

				String inVirtualHostName = null;
				String inURIPath = null;
				int firstSlash = uriString.indexOf("/");
				if (firstSlash >= 0)
				{
					inVirtualHostName = uriString.substring(0, firstSlash);
					inURIPath = uriString.substring(firstSlash);
				}
				else
				{
					RemoteException r = new RemoteException(nls.getFormattedMessage("sb.seturis.bad.uri1.exception",
																																					new Object[] { uriString},
																																					"Incorrect input URI path format (path={0})."));
					if (tc.isEntryEnabled()) Tr.exit(tc, "setURIs", r);
					throw r;
				}

				// Determine of the given URI already exists.
				try
				{
					newURI = uriHome.findByName(uriString, false);

					// If the input URI was found, and this is not a Servlet Clone,
					// then determine if it already belongs to a Servlet or ServletGroup. 
					// If it does, then throw the URI already exsist exception. Else, just
					// use the unused URI.
					if (!isCloneInstance)
					{
						try
						{
							servletList = newURI.getAllServlets();
						}
						catch (Exception e)
						{
						}

						// If the URI found is NOT already associated with this servlet,
						// then throw the URI already exists exception.
						try
						{
							groupList = newURI.getAllServletGroups();
						}
						catch (Exception e)
						{
						}

						Node thisNode = null;
						Long thisNodeKey = null;
						Node servletNode = null;
						Long servletNodeKey = null;
						Node servletGroupNode = null;
						Long servletGroupNodeKey = null;

						boolean uriExists = false;
						InvalidURINameException iune= null;
						String fullName = null;

						if (!servletList.isEmpty())
						{

							listSize = servletList.size();
							for (int s = 0; s < listSize; s++)
							{
								servlet = (Servlet) servletList.elementAt(s);
								servletNode = servlet.getNode();
								servletNodeKey = (Long)(servletNode.getPrimaryKey());

								if (!hasContext)
								{
									servletGroup = initialServletGroup;
									thisNode = servletGroup.getNode();
									thisNodeKey = (Long)(thisNode.getPrimaryKey());

									// If the Servlet found for the URI is on the same Node that "this" Servlet 
									// is, then throw exception.
									if ((servletNodeKey.intValue()) == (thisNodeKey.intValue()))
									{
										uriExists = true;
										fullName = (servlet.getFullName()).toString();
										iune = new InvalidURINameException(nls.getFormattedMessage("sb.seturis.uri.already.exists1.exception",
																																							 new Object[] { fullName},
																																							 "URI specified already exists for resource (Servlet={0})."));
									}
								}
								else
								{
									thisNode = getNode();
									thisNodeKey = (Long)(thisNode.getPrimaryKey());

									primaryKey = (Long)(ec.getPrimaryKey());
									Long servletKey = (Long)(servlet.getPrimaryKey());

									// If the Servlet found for the URI is on the same Node that "this" Servlet 
									// is AND it is not looking itself, then throw exception.
									if (((servletNodeKey.intValue()) == (thisNodeKey.intValue())) &&
											((primaryKey.intValue()) != (servletKey.intValue())))
									{
										uriExists = true;
										fullName = (servlet.getFullName()).toString();
										iune = new InvalidURINameException(nls.getFormattedMessage("sb.seturis.uri.already.exists1.exception",
																																							 new Object[] { fullName},
																																							 "URI specified already exists for resource (Servlet={0})."));
									}
								}
							}
						}
						else if (!groupList.isEmpty())
						{
							listSize = groupList.size();
							for (int s = 0; s < listSize; s++)
							{
								servletGroup = (ServletGroup) groupList.elementAt(s);
								servletGroupNode = servletGroup.getNode();
								servletGroupNodeKey = (Long)(servletGroupNode.getPrimaryKey());

								if (!hasContext)
								{
									servletGroup = initialServletGroup;
									thisNode = servletGroup.getNode();
									thisNodeKey = (Long)(thisNode.getPrimaryKey());
								}
								else
								{
									thisNode = getNode();
									thisNodeKey = (Long)(thisNode.getPrimaryKey());
								}

								// If a ServletGroup was found for the URI and it is on the same Node that "this" Servlet 
								// is, then throw exception.
								if ((servletGroupNodeKey.intValue()) == (thisNodeKey.intValue()))
								{
									// Special Case:
									//  If the new URI for this servlet happens to be '/', then we must
									//  allow this URI to have a double meaning. One as the Root URI AND
									//  the other as a Servlet URI.
									if (!(inURIPath.equals("/")))
									{
									  uriExists = true;
									  fullName = (servletGroup.getFullName()).toString();
									  iune = new InvalidURINameException(nls.getFormattedMessage("sb.seturis.uri.already.exists2.exception",
										  																												 new Object[] { fullName},
																																						   "URI specified already exists for resource (Web Application={0})."));
									}
								}
							}
						}

						if (uriExists)
						{
							if (tc.isEntryEnabled()) Tr.exit(tc, "setURIs", iune);
							throw iune;
						}
					}
				}
				catch (ObjectNotFoundException e)
				{
					newURI = null;
				}

				// Get this Servlet's ServletGroup.
				try
				{
					String groupURI = null; 

					if (!hasContext)
					{
						servletGroup = initialServletGroup;
					}
					else
					{
						servletGroup = (ServletGroup) getContainingObject();
					}

					ServletGroupAttributes groupAttrs = new ServletGroupAttributes();
					groupAttrs = (ServletGroupAttributes) (servletGroup.getAttributes(groupAttrs));
					groupURI = groupAttrs.getURIPath();
					firstSlash = groupURI.indexOf("/");
					if (firstSlash >= 0)
					{
						groupPath = groupURI.substring(firstSlash);
					}
					else
					{
						RemoteException r = new RemoteException(nls.getFormattedMessage("sb.seturis.bad.uri2.exception",
																																						new Object[] { groupURI},
																																						"Incorrect ServletGroup URI path format (path={0})."));
						if (tc.isEntryEnabled()) Tr.exit(tc, "setURIs", r);
						throw r;
					}
				}
				catch (ObjectNotFoundException e)
				{
					RemoteException r = new RemoteException((nls.getString("sb.servlet.exception",
																																 "Servlet does not belong to a ServletGroup.")),
																									e);
					if (tc.isEntryEnabled()) Tr.exit(tc, "setURIs");
					throw r;
				}

				URIAttributes uriAttributes = new URIAttributes();

				// If the given URI was not found, create it now!
				if (newURI == null)
				{
					uriAttributes.setName(inURIPath);
					uriAttributes.setRootURI(groupPath);

					// Get the the parent ServletGroup's VirtualHost.
					VirtualHostHome virtualHostHome = (VirtualHostHome)javax.rmi.PortableRemoteObject.narrow(getHome("VirtualHostHome"), 
																																																	 VirtualHostHome.class);

					VirtualHost virtualHost = null;

					try
					{
						virtualHost = virtualHostHome.findByName(inVirtualHostName, false);
					}
					catch (ObjectNotFoundException e)
					{
						ObjectNotFoundException o = new ObjectNotFoundException(nls.getFormattedMessage("sb.seturis.virtualhost.not.found.exception",
																																														new Object[] { inVirtualHostName},
																																														"VirtualHost not found (name={0})."));
						if (tc.isEntryEnabled()) Tr.exit(tc, "setURIs", o);
						throw o;
					}

					try
					{
						newURI = uriHome.create(uriAttributes, virtualHost);
					}
					catch (CreateException e)
					{
						RemoteException r =  new RemoteException((nls.getString("sb.seturis.uri.create1.exception",
																																		"Exception occurred creating URI(s) instances for input paths.")), 
																										 e);
						if (tc.isEntryEnabled()) Tr.exit(tc, "setURIs", r);
						throw r;
					}
				}
				else
				{
					// Found an eligible URI, so refresh it's attributes!
					uriAttributes = (URIAttributes)(newURI.getAttributes(uriAttributes));
					uriAttributes.setPathName(inURIPath);
					uriAttributes.setRootURI(groupPath);
					newURI.setAttributes(uriAttributes);
				}

				if (newURI != null)
				{
					uris.addElement(newURI);
				}
				else
				{
					RemoteException r = new RemoteException(nls.getFormattedMessage("sb.seturis.uri.create2.exception",
																																					new Object[] { uriString},
																																					"Creation of URI for input path failed (path={0})."));
					if (tc.isEntryEnabled()) Tr.exit(tc, "setURIs", r);
					throw r;
				}

				// Add URI path to the developing Vector. This vector will contain paths
				// that don't have double slashes.
				uriPaths.addElement(uriString);
			}

			// Case 1: Initialize with Vector of URIs and set this Servlet's URI paths
			//         instance variable.
			if (!hasContext)
			{
				servletURIs = uriPaths;
				if (tc.isEntryEnabled()) Tr.exit(tc, "setURIs - Case 1");
				return;
			}

			// Case 2: Update current Vector of URIs with new input Vector. 
			if (hasContext)
			{
				primaryKey = (Long) ec.getPrimaryKey();

				currentURIs = listServletURIObjects();

				if (currentURIs == null)
				{
					ObjectNotFoundException o = new ObjectNotFoundException(nls.getFormattedMessage("sb.seturis.uri.not.found.exception",
																																													new Object[] { primaryKey},
																																													"No URI(s) for Servlet were found (id={0})."));
					if (tc.isEntryEnabled()) Tr.exit(tc, "setURI", o);
					throw o;
				}

				// Build Key list of current URIs. 
				uri = null;
				currentLen = currentURIs.size();
				Long currentKeyList[] = new Long[currentLen];

				for (int i = 0; i < currentLen; i++)
				{
					uri = (URI)(currentURIs.elementAt(i));
					currentKeyList[i] = (Long)(uri.getPrimaryKey());
				}

				// Build Key list of new URIs. 
				newLen = uris.size();
				Long newKeyList[] = new Long[newLen];

				for (int i = 0; i < newLen; i++)
				{
					uri = (URI)(uris.elementAt(i));
					newKeyList[i] = (Long)(uri.getPrimaryKey());
				}
             //@88536 Start of code

            if (isCloneInstance) {
                deleteFlag=ServletBeanHelper.getDeleteFlag((Long)ec.getPrimaryKey(),this);

            }//end if (isCloneInstance)
            //@88536 End of code

				// Loop thru current URIs to check for URI deletion. 
				boolean newURIs = false;
				j = 0;
				found = false;      

				for (int i = 0; i < currentLen; i++)
				{
					j = 0;
					found = false;      
					while ((j < newLen) && !found)
					{
						if ((currentKeyList[i].intValue()) == (newKeyList[j].intValue()))
						{
							found = true;
						}
						j++;
					}

					//*   @88536 CODE REMOVED  *********  start
               // Remove the URI only if this Servlet is NOT a Clone!
               //if ((!found) && (!isCloneInstance)) {
               //*   @88536 CODE REMOVED  *********  end
               if ((!isCloneInstance || (deleteFlag)) && (!found)) {	//@88536C					// Remove the old URI associated with this Servlet. 
						newURIs = true;
						uri = (URI)(currentURIs.elementAt(i));
						uri.setPrivateRemove(new Boolean(true));
						uri.remove();
					}
				}

				// Loop thru new URIs to check for URI addition. 
				j = 0;
				found = false;      

				for (int i = 0; i < newLen; i++)
				{
					j = 0;
					found = false;      
					while ((j < currentLen) && !found)
					{
						if ((newKeyList[i].intValue()) == (currentKeyList[j].intValue()))
						{
							found = true;
						}
						j++;
					}

					if (!found)
					{
						newURIs = true;
					}
				}

				if (newURIs)
				{
					// Persist the new servletURIs value. 
					servletURIs = uriPaths;
					setBinaryAttr(servletURIsBinAttrName, servletURIs);
					dirty = true;
					ejbStore();
				}

				if (tc.isEntryEnabled()) Tr.exit(tc, "setURIs - Case 2");
				return;
			}
		}
		catch (OpException e)
		{
			e.printStackTrace();
			RemoteException r =  new RemoteException((nls.getString("sb.seturis.uri.create.delete.exception",
																															"Exception occurred creating/deleting a URI instance.")), 
																							 e);
			if (tc.isEntryEnabled()) Tr.exit(tc, "setURIs", r);
			throw r;
		}
		catch (FinderException e)
		{
			RemoteException r = new RemoteException((nls.getString("sb.seturis.uri.find.exception",
																														 "Exception occurred attempting to find a URI path.")),
																							e);
			if (tc.isEntryEnabled()) Tr.exit(tc, "setURIs", r);
			throw r;
		}
		catch (RemoveException e)
		{
			RemoteException r = new RemoteException((nls.getString("sb.seturis.uri.remove.exception",
																														 "Exception occurred attempting to remove a URI instance.")),
																							e);
			if (tc.isEntryEnabled()) Tr.exit(tc, "setURIs", r);
			throw r;
		}
		catch (RemoteException e)
		{
			RemoteException r =  new RemoteException((nls.getString("sb.seturis.uri.create.delete.exception",
																															"Exception occurred creating/deleting a URI instance.")), 
																							 e);
			if (tc.isEntryEnabled()) Tr.exit(tc, "setURIs", r);
			throw r;
		}
		catch (AttributeNotSetException e)
		{
			RemoteException r =  new RemoteException((nls.getString("sb.seturis.uri.create.delete.exception",
																															"Exception occurred creating/deleting a URI instance.")), 
																							 e);
			if (tc.isEntryEnabled()) Tr.exit(tc, "setURIs", r);
			throw r;
		}

		if (tc.isEntryEnabled()) Tr.exit(tc, "setURIs");
	}

	/*
	 * renameURIsVirtualHost -- Renames all the fully-qualified URI paths for the Servlet
	 *                          from the old VirtualHost name to the new input VirtualHost name.
	 */
	public void renameURIsVirtualHost(String oldVirtualHostName, String newVirtualHostName)
	throws RemoteException
	{
		if (tc.isEntryEnabled()) Tr.entry(tc, "renameURIsVirtualHost");

		int numURIs = servletURIs.size();
		Vector newURIs = new Vector(numURIs);

		for (int i = 0; i < numURIs; i++)
		{
			String currentURI = (String)(servletURIs.elementAt(i));
			String currentVirtualHostName = null;
			String currentURIPath = null;
			String newURI = null;

			int firstSlash = currentURI.indexOf("/");
			if (firstSlash >= 0)
			{
				currentVirtualHostName = currentURI.substring(0, firstSlash);
				currentURIPath = currentURI.substring(firstSlash);
			}
			else
			{
				RemoteException r = new RemoteException(nls.getFormattedMessage("sb.renameurivh.bad.uri.exception",
																																				new Object[] { currentURI},
																																				"Incorrect URI path format found (path={0})."));
				if (tc.isEntryEnabled()) Tr.exit(tc, "renameURIsVirtualHost", r);
				throw r;
			}

			if (currentVirtualHostName.equals(oldVirtualHostName))
			{
				newURI = newVirtualHostName + currentURIPath;
			}
			else
			{
				newURI = currentURI;
			}

			newURIs.addElement(newURI);
		}

		// Persist the new servletURIs value. 
		servletURIs = newURIs;
		setBinaryAttr(servletURIsBinAttrName, servletURIs);
		dirty = true;
		ejbStore();

		if (tc.isEntryEnabled()) Tr.exit(tc, "renameURIsVirtualHost");
	}

	/*
	 * Helper Methods
	 */

	/*
	 * initializeNLS -- Creates and initializes the NLS object.
	 */
	protected static void initializeNLS(String propertiesFile)
	{
		if (tc.isEntryEnabled()) Tr.entry(tc, "initializeNLS");

		synchronized (classLock) 
		{
			if (nls != null)
			{
				if (tc.isEntryEnabled()) Tr.exit(tc, "initializeNLS");
				return;
			}
			else
			{
				nls = new NLS(propertiesFile);
			}
		}

		if (tc.isEntryEnabled()) Tr.exit(tc, "initializeNLS");
	}
}
